"""
Pytest automatically parses this module for fixtures that can be used in its
directory, as well as in any of its subdirectories.

This module contains fixtures that return generic NXfields.
"""

from pytest import fixture
import numpy as np

from nexusformat.nexus.tree import NXfield


@fixture
def x():
    return NXfield(2 * np.linspace(0.0, 10, 11, dtype=np.float64), name="x")


@fixture
def y():
    return NXfield(3 * np.linspace(0.0, 5.0, 6, dtype=np.float64), name="y")


@fixture
def z():
    return NXfield(4 * np.linspace(0.0, 2.0, 3, dtype=np.float64), name="z")


@fixture
def field1():
    return NXfield((1, 2), name="f1")


@fixture
def field2():
    return NXfield((3, 4), name="f2")


@fixture
def field3():
    return NXfield((5, 6), name="f3")


@fixture
def field4():
    return NXfield("a", name="f1")


@fixture
def arr1D():
    return np.linspace(0.0, 100.0, 101, dtype=np.float64)


@fixture
def arr2D():
    return np.array(((1, 2, 3, 4), (5, 6, 7, 8)), dtype=np.int32)


@fixture
def arr3D():
    return np.resize(np.linspace(0.0, 124.0, 125, dtype=np.float64), (5, 5, 5))


@fixture
def v():
    v = NXfield(np.linspace(0, 99, num=100, dtype=np.float64), name="v")
    v.resize((2, 5, 10))
    return v


@fixture
def im():
    return NXfield(np.ones(shape=(10, 10, 4), dtype=np.float32), name="image")



import h5py as h5
import numpy as np
import pytest
from nexusformat.nexus.tree import NXfield, nxgetencoding


@pytest.fixture
def string_dtype():
    return h5.special_dtype(vlen=str)


@pytest.fixture
def field():
    return NXfield(shape=(10, 5, 5), dtype=np.int16, maxshape=(20, 10, 10),
                   fillvalue=0)


@pytest.mark.parametrize("text", ["a", "abc", "αβγ"])
def test_string_field_creation(text, string_dtype):

    field = NXfield(text)

    assert field.nxvalue == text
    assert field.dtype == string_dtype
    assert field.is_string()
    assert len(field) == len(text)


@pytest.mark.parametrize("text", ["a", "abc", "αβγ"])
def test_byte_field_creation(text, string_dtype):

    field = NXfield(text, dtype='S')

    assert field.nxvalue == text
    assert field.nxdata.decode(nxgetencoding()) == text
    assert field.dtype != string_dtype
    assert field.is_string()
    assert len(field) == len(text)


@pytest.mark.parametrize(
    "arr",
    ["arr1D",
     "arr2D",
     "arr3D"])
def test_array_field_creation(arr, request):

    arr = request.getfixturevalue(arr)
    field = NXfield(arr)

    assert np.all(field.nxvalue == arr)
    assert np.all(field.nxdata == arr)
    assert field.shape == arr.shape
    assert field.dtype == arr.dtype
    assert field.size == arr.size
    assert field.is_numeric()
    assert len(field) == len(arr)
    assert field.reshape((field.size)) == NXfield(arr.reshape((arr.size)))


@pytest.mark.parametrize(
    "arr",
    ["arr1D",
     "arr2D",
     "arr3D"])
def test_binary_field_operations(arr, request):

    arr = request.getfixturevalue(arr)
    field = NXfield(arr)

    assert np.all((field+2).nxvalue == arr+2)
    assert np.all((field-2).nxvalue == arr-2)
    assert np.all((2*field).nxvalue == 2*arr)


@pytest.mark.parametrize(
    "arr",
    ["arr1D",
     "arr2D",
     "arr3D"])
def test_field_methods(arr, request):

    arr = request.getfixturevalue(arr)
    field = NXfield(arr)

    assert field.sum() == np.sum(arr)


@pytest.mark.parametrize(
    "arr,idx", [("arr1D", np.s_[2:5]),
                ("arr2D", np.s_[2:5, 2:5]),
                ("arr3D", np.s_[2:5, 2:5, 2:5])])
def test_field_slice(arr, idx, request):

    arr = request.getfixturevalue(arr)
    field = NXfield(arr)

    assert np.array_equal(field[idx].nxvalue, arr[idx])
    assert field[idx].shape == arr[idx].shape


def test_field_index(arr1D):

    field = NXfield(2*arr1D)

    assert field.index(10.) == 5
    assert field.index(11.) == 5
    assert field.index(11., max=True) == 6
    assert field.index(12., max=True) == 6

    field = NXfield(2*arr1D[::-1])

    assert field.index(10.) == 95
    assert field.index(11.) == 94
    assert field.index(11., max=True) == 95
    assert field.index(12., max=True) == 94


def test_field_resize(field):

    field[9] = 1

    assert field.shape == (10, 5, 5)
    assert field.sum() == 25

    field.resize((15, 5, 5))
    field[14] = 1

    assert field.shape == (15, 5, 5)
    assert field.sum() == 50

    field.resize((15, 5, 10))
    field[:, :, 9] = 1

    assert field.shape == (15, 5, 10)
    assert field[:, :, 9].sum() == 75


def test_field_printing(arr1D):

    assert str(NXfield(arr1D)) == str(arr1D)
    assert f"{NXfield(arr1D)[10]:g}" == f"{arr1D[10]:g}"



import numpy as np
import pytest
from nexusformat.nexus.tree import NXdata, NXentry, NXsample


@pytest.fixture
def entry_1(x, y, z, v):
    """Simple entry."""
    entry = NXentry(NXdata(2*v, (z, y, x), name='d1'), NXsample(name='s1'))
    entry.attrs['default'] = 'd1'
    return entry


@pytest.fixture
def entry_2(x, y, z, v):
    """Simple entry, but different to entry_1"""
    return NXentry(NXdata(v, (z, y, x), name='d1'))


@pytest.fixture
def entry_3(entry_1, entry_2):
    """Simple linear combination of entries."""
    return entry_1 + entry_2


@pytest.fixture
def entry_4(entry_1, entry_2):
    """Second simple linear combination of entries"""
    return entry_1 - entry_2


def test_entry_operation_values(entry_3, entry_4, v):

    assert np.array_equal(entry_3.d1.nxsignal.nxvalue, 3*v)
    assert np.array_equal(entry_4.d1.nxsignal.nxvalue, v)


@pytest.mark.parametrize(
    "entry",
    ['entry_3', 'entry_4']
)
def test_entry_operations(entry, entry_1, request):

    entry = request.getfixturevalue(entry)
    assert 'd1' in entry
    assert 's1' in entry
    assert entry.attrs['default'] == 'd1'

    assert entry.d1.nxaxes == entry_1.d1.nxaxes
    assert entry.d1.nxsignal.nxname == "v"
    assert [axis.nxname for axis in entry.d1.nxaxes] == ["z", "y", "x"]


def test_plottable_data(entry_2):

    assert entry_2.is_plottable()
    assert entry_2.plottable_data is entry_2['d1']



import os

import pytest
from nexusformat.nexus.tree import (NXdata, NXentry, NXFile, NXroot, nxload,
                                    nxopen)


def test_file_creation(tmpdir):

    filename = os.path.join(tmpdir, "file.nxs")
    with NXFile(filename, "w") as f:
        assert f.mode == "rw"
        assert f.filename == filename
    assert not f.is_open()


def test_file_save(tmpdir, field1, field2):

    filename = os.path.join(tmpdir, "file.nxs")

    w1 = NXroot(NXentry())
    w1.entry.data = NXdata(field1, field2)
    w1.save(filename)

    assert os.path.exists(filename)
    assert not w1.nxfile.is_open()
    assert w1.nxfilename == filename
    assert w1.nxfilemode == "rw"

    w2 = nxload(filename)
    assert w2.nxfilename == filename
    assert w2.nxfilemode == "r"
    assert "entry/data/f1" in w2
    assert "entry/data/f2" in w2
    assert "signal" in w2["entry/data"].attrs
    assert "axes" in w2["entry/data"].attrs


@pytest.mark.parametrize("recursive", ["True", "False"])
def test_file_recursion(tmpdir, field1, field2, recursive):

    filename = os.path.join(tmpdir, "file.nxs")
    w1 = NXroot(NXentry())
    w1.entry.data = NXdata(field1, field2)
    w1.save(filename)

    w2 = nxload(filename, recursive=recursive)

    if not recursive:
        assert w2["entry"]._entries is None
        assert "entry/data" in w2
        assert w2["entry"]._entries is not None
        assert w2["entry/data"]._entries is None
        assert "entry/data/f1" in w2
        assert w2["entry/data"]._entries is not None
        assert w2["entry/data/f2"] == field2

    assert "entry/data/f1" in w2
    assert "entry/data/f2" in w2
    assert "signal" in w2["entry/data"].attrs
    assert "axes" in w2["entry/data"].attrs


def test_file_context_manager(tmpdir, field1, field2):

    filename = os.path.join(tmpdir, "file.nxs")

    with nxopen(filename, "w") as w1:
        w1["entry"] = NXentry()
        w1["entry/data"] = NXdata(field1, field2)
        assert w1.nxfilename == filename
        assert w1.nxfilemode == "rw"

    assert os.path.exists(filename)

    w2 = nxopen(filename)
    assert w2.nxfilename == filename
    assert w2.nxfilemode == "r"
    assert "entry/data/f1" in w2
    assert "entry/data/f2" in w2
    assert "signal" in w2["entry/data"].attrs
    assert "axes" in w2["entry/data"].attrs



#!/usr/bin/env python
# -----------------------------------------------------------------------------
# Copyright (c) 2019-2021, NeXpy Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING, distributed with this software.
# -----------------------------------------------------------------------------
import argparse

from nexusformat import __version__
from nexusformat.nexus import nxdir


def main():

    parser = argparse.ArgumentParser(
        description="Print a NeXus file tree")
    parser.add_argument('file', action='store', help="name of NeXus file")
    parser.add_argument('-s', '--short', action='store_true',
                        help="print only the first level")
    parser.add_argument('-v', '--version', action='version',
                        version=f'nxdir v{__version__}')

    args = parser.parse_args()

    nxdir(args.file, short=args.short)


if __name__ == "__main__":
    main()



# -----------------------------------------------------------------------------
# Copyright (c) 2013-2021, NeXpy Development Team.
#
# Author: Paul Kienzle, Ray Osborn
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING, distributed with this software.
# -----------------------------------------------------------------------------

"""Module to provide standard Matplotlib plotting to the NeXus Python API."""
import copy

import numpy as np

from . import NeXusError, NXfield


def centers(axis, dimlen):
    """Return the centers of the axis bins.

    This works regardless of whether the axis consists of bin boundaries,
    i.e, `dimlen = len(axis) + 1``, or centers, i.e., `dimlen = len(axis)`.

    Parameters
    ----------
    axis : ndarray
        Array containing the axis values.
    dimlen : int
        Length of corresponding data dimension.

    Returns
    -------
    ndarray
        Array of bin centers with a size of dimlen.
    """
    ax = axis.astype(np.float64)
    if ax.shape[0] == dimlen+1:
        return (ax[:-1] + ax[1:])/2
    else:
        assert ax.shape[0] == dimlen
        return ax


def boundaries(axis, dimlen):
    """Return the axis bin boundaries.

    This works regardless of whether the axis consists of bin boundaries,
    i.e, dimlen = len(axis) + 1, or centers, i.e., dimlen = len(axis).

    Parameters
    ----------
    axis : ndarray
        Array containing the axis values.
    dimlen : int
        Length of corresponding data dimension.

    Returns
    -------
    ndarray
        Array of bin boundaries with a size of dimlen + 1.
    """
    ax = axis.astype(np.float64)
    if ax.shape[0] == dimlen:
        start = ax[0] - (ax[1] - ax[0])/2
        end = ax[-1] + (ax[-1] - ax[-2])/2
        return np.concatenate((np.atleast_1d(start),
                               (ax[:-1] + ax[1:])/2,
                               np.atleast_1d(end)))
    else:
        assert ax.shape[0] == dimlen + 1
        return ax


def label(field):
    """Return a label for a data field suitable for use on a graph axis.

    This returns the attribute 'long_name' if it exists, or the field name,
    followed by the units attribute if it exists.

    Parameters
    ----------
    field : NXfield
        NeXus field used to construct the label.

    Returns
    -------
    str
        Axis label.
    """
    if 'long_name' in field.attrs:
        return field.long_name
    elif 'units' in field.attrs:
        return f"{field.nxname} ({field.units})"
    else:
        return field.nxname


class PylabPlotter(object):
    """Matplotlib plotter class for 1D or 2D NeXus data.

    When the nexusformat package is used within NeXpy, plots are produced by
    calling the NXPlotView class function, 'plot'. This provides a function
    with the same call signature for use outside NeXpy.
    """

    def plot(self, data_group, fmt=None, xmin=None, xmax=None,
             ymin=None, ymax=None, vmin=None, vmax=None, **kwargs):
        """Plot the NXdata group.

        Parameters
        ----------
        data_group : NXdata
            NeXus group containing the data to be plotted.
        fmt : str, optional
            Formatting options that are compliant with PyPlot, by default None
        xmin : float, optional
            Minimum x-boundary, by default None
        xmax : float, optional
            Maximum x-boundary, by default None
        ymin : float, optional
            Minimum y-boundary, by default None
        ymax : float, optional
            Maximum y-boundary, by default None
        vmin : float, optional
            Minimum signal value for 2D plots, by default None
        vmax : float, optional
            Maximum signal value for 2D plots, by default None
        **kwargs : dict
            Options used to customize the plot.

        Note
        ----
        If the qualitative color map, 'tab10', is specified as a
        keyword argument, the color scale is chosen so that all the
        integer values are centered on each color band, if the
        maximum intensity is less than 10.
        """
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise NeXusError(
                "Default plotting package (matplotlib) not available.")

        over = kwargs.pop("over", False)
        image = kwargs.pop("image", False)
        log = kwargs.pop("log", False)
        logx = kwargs.pop("logx", False)
        logy = kwargs.pop("logy", False)
        origin = kwargs.pop("origin", "lower")
        aspect = kwargs.pop("aspect", "auto")
        regular = kwargs.pop("regular", False)
        cmap = kwargs.pop("cmap", "viridis")
        colorbar = kwargs.pop("colorbar", True)
        interpolation = kwargs.pop("interpolation", "nearest")
        bad = kwargs.pop("bad", "darkgray")
        weights = kwargs.pop("weights", False)
        ax = kwargs.pop("ax", None)

        signal = data_group.nxsignal
        if signal.ndim > 2 and not image:
            raise NeXusError(
                "Can only plot 1D and 2D data - please select a slice")
        errors = data_group.nxerrors
        title = data_group.nxtitle

        # Provide a new view of the data if there is a dimension of length 1
        data, axes = (signal.nxdata.reshape(data_group.plot_shape),
                      data_group.plot_axes)

        if weights and data_group.nxweights:
            with np.errstate(divide='ignore'):
                w = data_group.nxweights.nxdata.reshape(data_group.plot_shape)
                data = np.where(w > 0, data/w, 0.0)

        isinteractive = plt.isinteractive()
        plt.ioff()

        try:
            if over:
                plt.autoscale(False)
            else:
                plt.autoscale(True)
                if ax is None:
                    plt.clf()
            if ax:
                plt.sca(ax)
            else:
                ax = plt.gca()

            # One-dimensional Plot
            if len(data.shape) == 1:
                if 'marker' in kwargs:
                    fmt = kwargs.pop('marker')
                else:
                    fmt = 'o'
                if 'units' in signal.attrs:
                    if not errors and signal.attrs['units'] == 'counts':
                        errors = NXfield(np.sqrt(data))
                if errors:
                    ebars = errors.nxdata
                    ax.errorbar(centers(axes[0], data.shape[0]), data, ebars,
                                fmt=fmt, **kwargs)
                else:
                    ax.plot(centers(axes[0], data.shape[0]), data, fmt,
                            **kwargs)
                if not over:
                    if xmin is not None:
                        ax.set_xlim(left=xmin)
                    if xmax is not None:
                        ax.set_xlim(right=xmax)
                    if ymin is not None:
                        ax.set_ylim(bottom=ymin)
                    if ymax is not None:
                        ax.set_ylim(top=ymax)
                    if logx:
                        ax.set_xscale('log')
                    if log or logy:
                        ax.set_yscale('log')
                    plt.xlabel(label(axes[0]))
                    plt.ylabel(label(signal))
                    plt.title(title)

            # Two dimensional plot
            else:
                from matplotlib.cm import get_cmap
                from matplotlib.colors import LogNorm, Normalize

                if image:
                    x = boundaries(axes[-1], data.shape[-2])
                    y = boundaries(axes[-2], data.shape[-3])
                    xlabel, ylabel = label(axes[-1]), label(axes[-2])
                else:
                    x = boundaries(axes[-1], data.shape[-1])
                    y = boundaries(axes[-2], data.shape[-2])
                    xlabel, ylabel = label(axes[-1]), label(axes[-2])

                if not vmin:
                    vmin = np.nanmin(data[data > -np.inf])
                if not vmax:
                    vmax = np.nanmax(data[data < np.inf])

                if image:
                    im = ax.imshow(data, origin='upper', **kwargs)
                    ax.set_aspect('equal')
                else:
                    if log:
                        vmin = max(vmin, 0.01)
                        vmax = max(vmax, 0.01)
                        kwargs["norm"] = LogNorm(vmin, vmax)
                    else:
                        kwargs["norm"] = Normalize(vmin, vmax)
                    cm = copy.copy(get_cmap(cmap))
                    cm.set_bad(bad, 1.0)
                    if regular:
                        extent = (x[0], x[-1], y[0], y[-1])
                        kwargs['interpolation'] = interpolation
                        im = ax.imshow(data, origin=origin, extent=extent,
                                       cmap=cm, **kwargs)
                    else:
                        im = ax.pcolormesh(x, y, data, cmap=cm, **kwargs)
                        ax.set_xlim(x[0], x[-1])
                        ax.set_ylim(y[0], y[-1])
                    if aspect == 'equal':
                        try:
                            if 'scaling_factor' in axes[-1].attrs:
                                _xscale = axes[-1].attrs['scaling_factor']
                            else:
                                _xscale = 1.0
                            if 'scaling_factor' in axes[-2].attrs:
                                _yscale = axes[-2].attrs['scaling_factor']
                            else:
                                _yscale = 1.0
                            aspect = float(_yscale / _xscale)
                        except Exception as error:
                            raise NeXusError(str(error))
                    ax.set_aspect(aspect)
                    if colorbar:
                        cb = plt.colorbar(im)
                        if cmap == 'tab10':
                            from matplotlib import __version__ as mpl_version
                            from pkg_resources import parse_version as pv
                            cmin, cmax = im.get_clim()
                            if cmax - cmin <= 9:
                                if cmin == 0:
                                    im.set_clim(-0.5, 9.5)
                                elif cmin == 1:
                                    im.set_clim(0.5, 10.5)
                                if pv(mpl_version) >= pv('3.5.0'):
                                    cb.ax.set_ylim(cmin-0.5, cmax+0.5)
                                    cb.set_ticks(range(int(cmin), int(cmax)+1))

                if xmin is not None:
                    ax.set_xlim(left=xmin)
                if xmax is not None:
                    ax.set_xlim(right=xmax)
                if ymin is not None:
                    if image:
                        ax.set_ylim(top=ymin)
                    else:
                        ax.set_ylim(bottom=ymin)
                if ymax is not None:
                    if image:
                        ax.set_ylim(bottom=ymax)
                    else:
                        ax.set_ylim(top=ymax)

                plt.xlabel(xlabel)
                plt.ylabel(ylabel)
                plt.title(title)

            if isinteractive:
                plt.pause(0.001)
                plt.show(block=False)
            else:
                plt.show()

        finally:
            if isinteractive:
                plt.ion()


plotview = PylabPlotter()



#!/usr/bin/env python
# -----------------------------------------------------------------------------
# Copyright (c) 2013-2021, NeXpy Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING, distributed with this software.
# -----------------------------------------------------------------------------
import argparse
import glob
import os
import re
import socket
import sys
import timeit

import numpy as np

from nexusformat.nexus import (NeXusError, NXdata, NXentry, NXfield, NXnote,
                               NXprocess, NXroot, nxsetcompression, nxversion)

prefix_pattern = re.compile(r'^([^.]+)(?:(?<!\d)|(?=_))')
index_pattern = re.compile(r'^(.*?)([0-9]*)[.](.*)$')


def natural_sort(key):
    import re
    return [int(t) if t.isdigit() else t for t in re.split(r'(\d+)', key)]


def get_prefixes(directory):
    prefixes = []
    for filename in os.listdir(directory):
        f = filename.split(os.path.extsep)[0]
        match = prefix_pattern.match(f)
        if match:
            prefixes.append(match.group(1).strip('-').strip('_'))
    return list(set(prefixes))


def get_files(directory, prefix, extension, first=None, last=None,
              reverse=False):
    if not extension.startswith('.'):
        extension = '.' + extension
    filenames = sorted(
        glob.glob(os.path.join(directory, prefix + '*' + extension)),
        key=natural_sort, reverse=reverse)
    if len(filenames) == 0:
        print("No filenames matched!")
        exit(0)
    max_index = get_index(filenames[-1])
    if first:
        min_index = first
    else:
        min_index = get_index(filenames[0])
    if last:
        max_index = last
    else:
        max_index = get_index(filenames[-1])
    return [filename for filename in filenames if
            min_index <= get_index(filename) <= max_index]


def get_index(filename):
    return int(index_pattern.match(filename).group(2))


def read_image(filename):
    try:
        import fabio
    except ImportError:
        raise NeXusError("Please install the 'fabio' module")
    im = fabio.open(filename)
    return im.data


def read_images(filenames, shape):
    good_files = [f for f in filenames if f is not None]
    if good_files:
        v0 = read_image(good_files[0])
        if v0.shape != shape:
            raise AssertionError(
                f'Image shape of {good_files[0]} not consistent')
        v = np.empty([len(filenames), v0.shape[0], v0.shape[1]],
                     dtype=np.float32)
    else:
        v = np.empty([len(filenames), shape[0], shape[1]], dtype=np.float32)
    v.fill(np.nan)
    for i, filename in enumerate(filenames):
        if filename:
            v[i] = read_image(filename)
    return v


def initialize_nexus_file(directory, output_file, filenames, first):
    z_size = get_index(filenames[-1]) - get_index(filenames[0]) + 1
    v0 = read_image(filenames[0])
    x = NXfield(range(v0.shape[1]), dtype=np.uint16, name='x_pixel')
    y = NXfield(range(v0.shape[0]), dtype=np.uint16, name='y_pixel')
    if z_size > 1:
        z = first + np.arange(z_size)
        z = NXfield(z, dtype=np.uint16, name='frame_number', maxshape=(5000,))
        v = NXfield(name='data', shape=(z_size, v0.shape[0], v0.shape[1]),
                    dtype=np.float32,
                    maxshape=(5000, v0.shape[0], v0.shape[1]))
        data = NXdata(v, (z, y, x))
    else:
        v = NXfield(name='data', shape=(v0.shape[0], v0.shape[1]),
                    dtype=np.float32)
        data = NXdata(v, (y, x))
    root = NXroot(NXentry(data))
    root.save(output_file, 'w')
    return root


def write_data(root, filenames):
    if len(root.entry.data.data.shape) == 2:
        root.entry.data.data[:, :] = read_image(filenames[0])
    else:
        z_size = root.entry.data.data.shape[0]
        image_shape = root.entry.data.data.shape[1:3]
        chunk_size = root.entry.data.data.chunks[0]
        min_index = get_index(filenames[0])
        max_index = get_index(filenames[-1])
        k = 0
        for i in range(min_index, min_index+z_size, chunk_size):
            files = []
            for j in range(i, min(i+chunk_size, max_index+1)):
                if j == get_index(filenames[k]):
                    print('Processing', filenames[k])
                    files.append(filenames[k])
                    k += 1
                elif k < len(filenames):
                    files.append(None)
                else:
                    break
            root.entry.data.data[i-min_index:i+len(files)-min_index, :, :] = (
                read_images(files, image_shape))


def main():

    parser = argparse.ArgumentParser(
        description="Stack images into a single NeXus file")
    parser.add_argument('-d', '--directory', default=os.getcwd(),
                        help="directory containing the raw images")
    parser.add_argument('-p', '--prefix', nargs='+',
                        help="common prefix to all images")
    parser.add_argument('-e', '--extension', default='.tif',
                        help="file extension of raw images")
    parser.add_argument('-o', '--output', help="name of NeXus output file")
    parser.add_argument('-f', '--first', default=0, type=int,
                        help="first frame to be included in the stacked data")
    parser.add_argument('-l', '--last', type=int,
                        help="last frame to be included in the stacked data")
    parser.add_argument('-r', '--reverse', action='store_true',
                        help="store images in reverse order")
    parser.add_argument('-c', '--compression', help="HDF5 compression method")
    parser.add_argument('-v', '--version', action='version',
                        version=f'nxstack v{nxversion}')

    args = parser.parse_args()
    directory = args.directory
    extension = args.extension
    if not extension.startswith('.'):
        extension = '.' + extension
    output = args.output
    if output is not None and os.path.splitext(output)[1] == '':
        output = output + '.nxs'

    compression = args.compression
    if compression:
        nxsetcompression(compression)

    first = args.first
    last = args.last
    reverse = args.reverse

    if args.prefix is None:
        prefixes = get_prefixes(directory)
        if len(prefixes) > 1 and output is not None:
            raise NeXusError(
                "Only one prefix allowed if the output file is specified")
    else:
        prefixes = args.prefix

    for prefix in prefixes:
        tic = timeit.default_timer()
        data_files = get_files(directory, prefix, extension, first, last,
                               reverse)
        if output is None:
            output_file = prefix + '.nxs'
        else:
            output_file = output
        root = initialize_nexus_file(directory, output_file, data_files, first)
        write_data(root, data_files)
        note = NXnote('nxstack '+' '.join(sys.argv[1:]),
                      (f'Current machine: {socket.gethostname()}\n'
                       f'Current working directory: {os.getcwd()}\n'
                       f'Data files: {data_files[0]} to {data_files[-1]}\n'))
        root.entry['nxstack'] = NXprocess(
            program='nxstack', sequence_index=len(root.entry.NXprocess) + 1,
            version=nxversion, note=note)

        toc = timeit.default_timer()
        print(toc-tic, 'seconds for', output_file)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -----------------------------------------------------------------------------
# Copyright (c) 2019-2021, NeXpy Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING, distributed with this software.
# -----------------------------------------------------------------------------
import argparse

from nexusformat import __version__
from nexusformat.nexus import nxduplicate


def main():

    parser = argparse.ArgumentParser(
        description="Copy a NeXus file to another file")
    parser.add_argument('input', action='store',
                        help="name of NeXus input file")
    parser.add_argument('output', action='store',
                        help="name of NeXus output file")
    parser.add_argument('-e', '--expand_external', action='store_true',
                        help="store external links within the new file")
    parser.add_argument('-o', '--overwrite', action='store_true',
                        help="overwrite any existing file")
    parser.add_argument('-v', '--version', action='version',
                        version=f'nxduplicate v{__version__}')

    args = parser.parse_args()

    if args.overwrite:
        mode = 'w'
    else:
        mode = 'w-'
    nxduplicate(args.input, args.output, mode=mode,
                expand_external=args.expand_external)


if __name__ == "__main__":
    main()



import os
import warnings

import numpy as np
import pytest
from nexusformat.nexus.tree import (NXdata, NXentry, NXfield, NXroot,
                                    NXsubentry, NXvirtualfield, nxconsolidate,
                                    nxload)


@pytest.fixture
def data(x, y, z, v):
    """
    Returns a simple NXdata object with a title.
    """
    return NXdata(v, (z, y, x), title="Title")


@pytest.fixture
def data2(v):
    """
    Returns a different, simple NXdata object with a title.
    """
    v2 = v[0]
    v2.resize((1, 5, 10))
    return NXdata(v2)


@pytest.fixture
def empty_data():
    """
    Returns a default NXdata instance.
    """
    return NXdata()


@pytest.fixture
def NXdata_from_empty_01(empty_data, x, y, z, v):
    empty_data.nxsignal = v
    empty_data.nxaxes = (z, y, x)
    return empty_data


@pytest.fixture
def NXdata_from_empty_02(empty_data, x, y, z, v):
    empty_data["v"] = v
    empty_data["x"] = x
    empty_data["y"] = y
    empty_data["z"] = z
    empty_data.nxsignal = "v"
    empty_data.nxaxes = ("z", "y", "x")
    return empty_data


def test_data_creation(data):

    assert "signal" in data.attrs
    assert "axes" in data.attrs
    assert len(data.attrs["axes"]) == 3
    assert data.ndim == 3
    assert data.shape == (2, 5, 10)
    assert data.nxsignal.nxname == "v"
    assert data.nxsignal.ndim == 3
    assert data.nxsignal.shape == (2, 5, 10)
    assert data.nxsignal.any()
    assert not data.nxsignal.all()
    assert [axis.nxname for axis in data.nxaxes] == ["z", "y", "x"]
    assert [axis.ndim for axis in data.nxaxes] == [1, 1, 1]
    assert [axis.shape for axis in data.nxaxes] == [(3,), (6,), (11,)]

    assert data.nxtitle == "Title"


def test_default_data(data):

    root = NXroot(NXentry(data))
    root["entry/data"].set_default()

    assert root.get_default() is root["entry/data"]
    assert root["entry"].get_default() is root["entry/data"]
    assert root["entry/data"].get_default() is root["entry/data"]
    assert root.plottable_data is root["entry/data"]

    root["entry/subentry"] = NXsubentry(data)
    root["entry/subentry/data"].set_default()

    assert root.get_default() is root["entry/data"]
    assert root["entry/subentry"].get_default() is root["entry/subentry/data"]
    assert root["entry/subentry"].plottable_data is root["entry/subentry/data"]

    root["entry/subentry/data"].set_default(over=True)

    assert root.get_default() is root["entry/subentry/data"]
    assert root["entry"].get_default() is root["entry/subentry/data"]
    assert root["entry/data"].get_default() is root["entry/data"]
    assert root.plottable_data is root["entry/subentry/data"]


def test_plottable_data_01(data):

    assert data.is_plottable()
    assert data.plottable_data is data
    assert data.plot_rank == 3
    assert data.plot_rank == data.nxsignal.ndim
    assert data.plot_axes == data.nxaxes
    assert data.nxsignal.valid_axes(data.nxaxes)


def test_plottable_data_02(data2):

    assert data2.shape == (1, 5, 10)
    assert data2.plot_shape == (5, 10)
    assert data2.plot_rank == 2
    assert data2.plot_rank == data2.nxsignal.ndim - 1


@pytest.mark.parametrize(
    "data_fixture_name",
    ['NXdata_from_empty_01',
     'NXdata_from_empty_02']
)
def test_signal_selection(data_fixture_name, z, v, request):

    data = request.getfixturevalue(data_fixture_name)
    assert data.nxsignal.nxname == "v"
    assert [axis.nxname for axis in data.nxaxes] == ["z", "y", "x"]
    assert np.array_equal(data.nxsignal, v)
    assert np.array_equal(data.nxaxes[0], z)


def test_rename(NXdata_from_empty_01):

    data = NXdata_from_empty_01
    data["x"].rename("xx")
    data["y"].rename("yy")
    data["z"].rename("zz")
    data["v"].rename("vv")
    assert data.nxsignal.nxname == "vv"
    assert [axis.nxname for axis in data.nxaxes] == ["zz", "yy", "xx"]


def test_size_one_axis(x, z):

    y1 = np.array((1), dtype=np.float64)
    v1 = NXfield(np.linspace(0, 10*1*2, num=10*1*2, dtype=np.int64), name="v")
    v1.resize((2, 1, 10))

    data = NXdata(v1, (z, y1, x))

    assert data.is_plottable()
    assert data.plottable_data is data
    assert data.plot_rank == 2
    assert data.plot_rank == data.nxsignal.ndim - 1
    assert len(data.plot_axes) == 2
    assert data.plot_shape == (2, 10)
    assert data.nxsignal.valid_axes(data.plot_axes)


def test_data_operations(data, v):

    new_data = data + 1

    assert np.array_equal(new_data.nxsignal.nxvalue, v + 1)
    assert new_data.nxaxes == data.nxaxes
    assert new_data.nxsignal.nxname == "v"
    assert [axis.nxname for axis in data.nxaxes] == ["z", "y", "x"]

    new_data = data - 1

    assert np.array_equal(new_data.nxsignal, v - 1)

    new_data = data * 2

    assert np.array_equal(new_data.nxsignal, v * 2)

    new_data = 2 * data

    assert np.array_equal(new_data.nxsignal, v * 2)

    new_data = data / 2

    assert np.array_equal(new_data.nxsignal, v / 2)

    new_data = 2 * data - data

    assert np.array_equal(new_data.nxsignal, v)


def test_data_errors():

    y1 = NXfield(np.linspace(1, 10, 10), name="y")
    v1 = NXfield(y1**2, name="v")
    e1 = NXfield(np.sqrt(v1))

    data = NXdata(v1, (y1), errors=e1)

    assert data.nxerrors is not None
    assert data.nxerrors.nxname == "v_errors"
    assert np.array_equal(data.nxerrors, e1)

    data = NXdata(v1, (y1))
    data.nxerrors = e1

    new_data = 2 * data

    assert np.array_equal(new_data.nxerrors, 2 * e1)

    new_data = 2 * data - data

    assert np.array_equal(new_data.nxerrors, e1 * np.sqrt(5))

    new_data = data - data / 2

    assert np.array_equal(new_data.nxerrors, e1 * np.sqrt(1.25))


def test_data_weights():

    y1 = NXfield(np.linspace(1, 10, 10), name="y")
    v1 = NXfield(y1**2, name="v")
    w1 = NXfield(np.sqrt(v1))

    data = NXdata(v1, (y1), errors=y1, weights=w1)

    assert data.nxweights is not None
    assert data.nxweights.nxname == "v_weights"
    assert np.array_equal(data.nxweights, w1)

    weighted_data = data.weighted_data()

    assert np.array_equal(weighted_data.nxsignal, v1 / w1)
    assert np.array_equal(weighted_data.nxerrors, y1 / w1)
    assert weighted_data.nxaxes == data.nxaxes

    data = NXdata(v1, (y1))
    data.nxweights = w1

    new_data = 2 * data

    assert np.array_equal(new_data.nxweights, 2 * w1)

    new_data = 2 * data - data

    assert np.array_equal(new_data.nxweights, w1)

    new_data = data - data / 2

    assert np.array_equal(new_data.nxweights, w1/2)


def test_data_angles(data):

    data.nxangles = [120, 90, 90]

    assert data.nxangles == [120.0, 90.0, 90.0]


def test_data_slabs(data, x, y, v):

    slab = data[0, :, :]

    assert np.array_equal(slab.nxsignal, v[0])
    assert slab.plot_rank == 2
    assert slab.plot_shape == v[0].shape
    assert slab.nxaxes == [y, x]

    slab = data[0, 3.:12., 2.:18.]

    assert slab.plot_shape == (v.shape[1]-2, v.shape[2]-2)
    assert slab.plot_axes == [y[1:-1], x[1:-1]]

    slab = data[0, 3.5:11.5, 2.5:17.5]

    assert slab.shape == (v.shape[1]-2, v.shape[2]-2)
    assert slab.plot_shape == (v.shape[1]-2, v.shape[2]-2)
    assert slab.plot_axes == [y[1:-1], x[1:-1]]

    slab1 = data[0:0, 3.5:11.5, 2.5:17.5]
    slab2 = data[0:1, 3.5:11.5, 2.5:17.5]

    assert slab1.shape == slab.shape
    assert slab2.shape == slab.shape


def test_data_projections(x, y, z, v):

    d1 = NXdata(v[0], (y, x))

    assert d1.nxaxes == [d1["y"], d1["x"]]

    p1 = d1.project((1, 0))
    p2 = d1.project((0, 1), limits=((3., 9.), (4., 16.)))

    assert p1.nxaxes == [p1["x"], p1["y"]]
    assert np.array_equal(p1["x"].nxvalue, d1["x"])
    assert p2.nxaxes == [p2["y"], p2["x"]]
    assert np.array_equal(p2["x"].nxvalue, d1["x"][4.:16.])
    assert np.array_equal(p2["x"].nxvalue, d1["x"][2:9])

    d2 = NXdata(v, (z, y, x))

    p3 = d2.project((0, 1), ((0., 8.), (3., 9.), (4., 16.)))

    assert p3.nxaxes == [p3["z"], p3["y"]]
    assert np.array_equal(p3["y"].nxvalue, d2["y"][3.:9.])
    assert np.array_equal(p3["y"].nxvalue, d2["y"][1:4])
    assert p3["x"] == 10.
    assert p3["x"].attrs["minimum"] == 4.
    assert p3["x"].attrs["maximum"] == 16.
    assert p3["x"].attrs["summed_bins"] == 7
    assert p3["v"].sum() == d2.v[:, 1:3, 2:8].sum()

    p4 = d2.project((0, 1), ((0., 8.), (3., 9.), (4., 16.)), summed=False)

    assert p4["v"].sum() == d2.v[:, 1:3, 2:8].sum() / \
        p4["x"].attrs["summed_bins"]


def test_data_smoothing(x):

    warnings.filterwarnings("ignore", message="numpy.ufunc size changed")
    data = NXdata(np.sin(x), (x))
    smooth_data = data.smooth(n=101, xmin=x.min(), xmax=x.max())

    assert smooth_data.nxsignal.shape == (101,)
    assert smooth_data.nxaxes[0].shape == (101,)
    assert smooth_data.nxsignal[0] == pytest.approx(np.sin(x)[0])
    assert smooth_data.nxsignal[-1] == pytest.approx(np.sin(x)[-1])

    smooth_data = data.smooth(factor=4)

    assert smooth_data.nxsignal.shape == (41,)
    assert smooth_data.nxaxes[0].shape == (41,)
    assert smooth_data.nxsignal[0] == pytest.approx(np.sin(x)[0])
    assert smooth_data.nxsignal[4] == pytest.approx(np.sin(x)[1])
    assert smooth_data.nxsignal[-1] == pytest.approx(np.sin(x)[-1])


def test_data_selection():

    xx = np.linspace(0, 20.0, 21, dtype=float)
    yy = np.ones(shape=xx.shape, dtype=float)
    yy[np.where(np.remainder(xx, 4) == 0.0)] = 2.0
    data = NXdata(yy, xx)

    selected_data = data.select(4.0)

    assert selected_data.shape == (6,)
    assert np.all(selected_data.nxsignal == 2.0)

    yy[(np.array((1, 3, 5, 7, 9, 11, 13, 15, 17, 19)),)] = 1.5
    data = NXdata(yy, xx)

    selected_data = data.select(4.0, offset=1.0)

    assert selected_data.shape == (5,)
    assert np.all(selected_data.nxsignal == 1.5)

    selected_data = data.select(4.0, offset=1.0, symmetric=True)

    assert selected_data.shape == (10,)
    assert np.all(selected_data.nxsignal == 1.5)


def test_image_data(x, y, z, v, im):

    root = NXroot(NXentry(NXdata(im)))
    root["entry"].attrs["default"] = "data"
    root["entry/other_data"] = NXdata(v, (z, y, x), title="Title")

    assert root["entry/data/image"].is_image()
    assert root["entry/data"].is_image()
    assert root.plottable_data.is_image()
    assert root["entry"].plottable_data.is_image()
    assert not root["entry/other_data"].is_image()


def test_smart_indices(x, v):

    ind = [1, 3, 5]

    assert all(x[ind].nxvalue == x.nxvalue[ind])
    assert all(v[v > 50].nxvalue == v.nxvalue[v.nxvalue > 50])
    assert all(v[1, 0, ind].nxvalue == v.nxvalue[1, 0, ind])

    x[ind] = 0

    assert x.any() and not x[ind].any()

    ind = np.array([[3, 7], [4, 5]])

    assert np.all(x[ind].nxvalue == x.nxvalue[ind])

    row = np.array([0, 1, 2])
    col = np.array([2, 1, 3])

    assert all(v[0][row, col].nxvalue == v[0].nxvalue[row, col])
    assert np.all(v[0][row[:, np.newaxis], col].nxvalue ==
                  v[0].nxvalue[row[:, np.newaxis], col])


@pytest.mark.parametrize("path", [True, False])
def test_virtual_fields(tmpdir, path, v):

    s1 = NXroot(NXentry(NXdata(v)))
    s2 = NXroot(NXentry(NXdata(2*v)))
    s3 = NXroot(NXentry(NXdata(3*v)))

    s1.save(os.path.join(tmpdir, "s1.nxs"), "w")
    s2.save(os.path.join(tmpdir, "s2.nxs"), "w")
    s3.save(os.path.join(tmpdir, "s3.nxs"), "w")

    sources = [f.nxfilename for f in [s1, s2, s3]]

    if path:
        vds1 = NXvirtualfield("entry/data/v", sources, shape=v.shape,
                              dtype=v.dtype)
    else:
        vds1 = NXvirtualfield(s1["entry/data/v"], sources)

    assert vds1.shape == (3,) + v.shape
    assert vds1.dtype == v.dtype
    assert vds1.sum() == 6 * v.sum()

    vds2 = nxconsolidate(sources, "entry/data")

    assert vds2.nxsignal.shape == vds1.shape
    assert vds2.nxsignal.dtype == v.dtype
    assert vds2.sum() == 6 * v.sum()

    NXroot(NXentry(vds2)).save(os.path.join(tmpdir, "vds.nxs"), "w")
    vds3 = nxload(os.path.join(tmpdir, "vds.nxs"))

    assert "entry/data/v" in vds3
    assert vds3["entry/data/v"].shape == vds1.shape
    assert vds3["entry/data/v"].dtype == v.dtype
    assert vds3["entry/data/v"].sum() == 6 * v.sum()
    assert vds3.nxfile["entry/data/v"].is_virtual



"""
This package provides a Python API that subclasses the nexusformat
package to work with HDF5 files served by h5serv.

"""

import h5pyd as h5
import numpy as np
from nexusformat.nexus import NXfield, NXFile, NXgroup, NXlinkfield

NX_SERVER = 'http://hdfgroup.org:5000'
NX_DOMAIN = 'exfac.org'

__all__ = ['NXRemoteFile', 'nxloadremote',
           'nxgetserver', 'nxsetserver', 'NX_SERVER',
           'nxgetdomain', 'nxsetdomain', 'NX_DOMAIN']


class NXRemoteFile(NXFile):

    """
    Structure-based interface to the NeXus file API.

    Usage::

      file = NXFile(filename, ['r','rw','w'])
        - open the NeXus file
      root = file.readfile()
        - read the structure of the NeXus file.  This returns a NeXus tree.
      file.writefile(root)
        - write a NeXus tree to the file.

    Example::

      nx = NXFile('REF_L_1346.nxs','r')
      tree = nx.readfile()
      for entry in tree.NXentry:
          process(entry)
      copy = NXFile('modified.nxs','w')
      copy.writefile(tree)

    Note that the large datasets are not loaded immediately.  Instead, the
    when the data set is requested, the file is reopened, the data read, and
    the file closed again.  open/close are available for when we want to
    read/write slabs without the overhead of moving the file cursor each time.
    The :class:`NXdata` objects in the returned tree hold the object values.
    """

    def __init__(self, name, mode='r', server=None,  domain=None,
                 **kwds):
        """
        Creates an h5py File object for reading and writing.
        """
        self.h5 = h5
        self.name = name
        self._mode = 'r'
        if server is None:
            server = NX_SERVER
        self._server = server
        if domain is None:
            domain = NX_DOMAIN
        self._domain = domain
        self._file = self.h5.File(self.domain, mode, endpoint=server)
        self._filename = self.domain
        self._path = '/'

    def __repr__(self):
        return f'<NXRemoteFile "{self._filename}" (mode {self._mode})>'

    def open(self, **kwds):
        if not self.isopen():
            self._file = self.h5.File(self._filename, self._mode,
                                      endpoint=self._server)
            self.nxpath = '/'
        return self

    def close(self):
        if self.isopen():
            self._file.close()

    def isopen(self):
        if self._file.id.uuid != 0:
            return True
        else:
            return False

    @property
    def domain(self):
        domain = self.name.split('.')[0].split('/')
        domain.reverse()
        domain.append(self._domain)
        return '.'.join(domain)

    @property
    def file(self):
        if not self.isopen():
            self.open()
        return self._file

    def readfile(self):
        """
        Reads the NeXus file structure from the file and returns a tree
        of NXobjects.

        Large datasets are not read until they are needed.
        """
        self.nxpath = '/'
        root = self._readgroup('root', self._file['/'])
        root._group = None
        root._file = self
        root._filename = self._filename
        root._mode = self._mode
        return root

    def _readchildren(self, group):
        children = {}
        for name, value in group.items():
            self.nxpath = self.nxpath + '/' + name
            if isinstance(value, self.h5.Group):
                children[name] = self._readgroup(name, value)
            else:
                children[name] = self._readdata(name, value)
            self.nxpath = self.nxparent
        return children

    def _readgroup(self, name, group):
        """
        Reads the group with the current path and returns it as an NXgroup.
        """
        if group is None:
            return NXgroup(name=name)
        attrs = self._readattrs()
        nxclass = self._readnxclass(attrs)
        if nxclass is not None:
            del attrs['NX_class']
        elif self.nxpath == '/':
            nxclass = 'NXroot'
        else:
            nxclass = 'NXgroup'
        children = self._readchildren(group)
        group = NXgroup(nxclass=nxclass, name=name, attrs=attrs,
                        entries=children)
        for obj in children.values():
            obj._group = group
        group._changed = True
        return group

    def _readdata(self, name, field):
        """
        Reads a data object and returns it as an NXfield or NXlink.
        """
        if field is None:
            return NXfield(name=name)
        value, shape, dtype, attrs = self.readvalues(field)
        if 'target' in attrs and self.nxpath != 'target':
            return NXlinkfield(
                value=value, name=name, dtype=dtype, shape=shape,
                target=self.attrs['target'],
                attrs=attrs)
        else:
            return NXfield(value=value, name=name, dtype=dtype, shape=shape,
                           attrs=attrs)

    def _readattrs(self):
        return dict(self[self.nxpath].attrs.items())

    def readvalues(self, field):
        shape, dtype = field.shape, field.dtype
        if shape == (1,):
            shape = ()
        # Read in the data if it's not too large
        if np.prod(shape) < 1000:  # i.e., less than 1k dims
            try:
                value = field[()]
                if isinstance(value, np.ndarray) and value.shape == (1,):
                    value = np.asscalar(value)
            except ValueError:
                value = None
        else:
            value = None
        attrs = self.attrs
        return value, shape, dtype, attrs


def getserver():
    global NX_SERVER
    return NX_SERVER


nxgetserver = getserver


def setserver(value):
    """
    Sets the default server and port for remote access.
    """
    global NX_SERVER
    NX_SERVER = value


nxsetserver = setserver


def getdomain():
    global NX_DOMAIN
    return NX_DOMAIN


nxgetdomain = getdomain


def setdomain(value):
    """
    Sets the default domain for remote access.
    """
    global NX_DOMAIN
    NX_DOMAIN = value


nxsetdomain = setdomain


def loadremote(filename, server=None, domain=None, mode='r'):
    """
    Reads a remote NeXus file returning a tree of objects.
    """
    if server is None:
        server = NX_SERVER
    if domain is None:
        domain = NX_DOMAIN
    with NXRemoteFile(filename, mode, server=server, domain=domain) as f:
        tree = f.readfile()
    return tree


nxloadremote = loadremote



# -----------------------------------------------------------------------------
# Copyright (c) 2013-2021, NeXpy Development Team.
#
# Author: Paul Kienzle, Ray Osborn
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING, distributed with this software.
# -----------------------------------------------------------------------------

"""
Python NeXus interface.

NeXus is a common data format for neutron, Xray and muon science.
The files contain multidimensional data elements grouped into a
hierarchical structure.  The data sets are self-describing, with
a description of the instrument configuration including the units
used as well as the data measured. NeXus data are written to HDF
files and accessed in this package using h5py.

Example
=======

First we need to load the file structure::

    import nexusformat.nexus as nx
    f = nx.load('file.nxs')

We can examine the file structure using a number of commands::

    f.attrs             # Shows file name, date, user, and NeXus version
    print(f.tree)       # Lists the entire contents of the NeXus file
    f.NXentry           # Shows the list of scans in the file
    f.NXentry[0].dir()  # Lists the fields in the first entry

Some files can even be plotted automatically::

    f.NXentry[0].NXdata[0].plot()


For a complete description of the features available in this tree view
of the NeXus data file, see `nx.tree`.
"""
from .completer import nxcompleter
from .lock import NXLock, NXLockException
from .tree import *



from nexusformat.nexus.tree import NXentry, NXgroup, NXroot


def test_attribute_paths(field1):

    root = NXroot(NXentry())
    root.entry.g1 = NXgroup(field1)

    assert root.entry.g1.nxpath == "/entry/g1"
    assert root["entry/g1"] is root.entry.g1
    assert root["entry/g1/f1"] is root.entry.g1.f1
    assert "g1" in root.entry
    assert "f1" in root.entry.g1
    assert "entry/g1/f1" in root
    assert root.entry.g1.f1.nxroot is root


def test_dictionary_paths(field1):

    root = NXroot(NXentry())
    root["entry/g1"] = NXgroup(field1)

    assert root.entry.g1.nxpath == "/entry/g1"
    assert root["entry/g1"] is root.entry.g1
    assert root["entry/g1/f1"] is root.entry.g1.f1
    assert "g1" in root["/entry"]
    assert "f1" in root["/entry/g1"]
    assert "/entry/g1/f1" in root
    assert root["/entry/g1/f1"].nxroot is root


def test_relative_paths(field1):

    root = NXroot(NXentry())
    root["entry/g1"] = NXgroup()
    root["entry/g1/g2"] = NXgroup()
    root["entry/g1/g2/f1"] = field1

    assert "f1" in root["entry/g1/g2"]
    assert "g2/f1" in root["entry/g1"]
    assert "g1/g2/f1" in root["entry"]
    assert root["entry/g1/g2/f1"].nxpath == "/entry/g1/g2/f1"
    assert "entry" in root



import os

from nexusformat.nexus.tree import NXdata, NXentry, NXgroup, NXlink, NXroot


def test_group_creation(field1, field2, field3):

    group1 = NXgroup(name="group")

    assert group1
    assert len(group1) == 0
    assert group1.nxname == "group"
    assert group1.nxgroup is None

    group2 = NXgroup(field1)

    assert group2
    assert len(group2) == 1
    assert "f1" in group2

    group1["f2"] = field2

    assert "f2" in group1

    group1["g2"] = group2

    assert len(group1) == 2
    assert "g2" in group1
    assert group1["g2/f1"] is group1.g2.f1

    group1["g2/f3"] = field3

    assert "f3" in group1["g2"]
    assert "g2/f3" in group1
    assert group1["g2/f3"].nxgroup == group1["g2"]

    group3 = NXgroup(g1=group1)

    assert "g1/g2/f1" in group3
    assert group3["g1"].nxgroup == group3


def test_group_insertion(field2):

    group1 = NXgroup()

    group1.insert(field2, name="f1")

    assert "f1" in group1
    assert len(group1) == 1


def test_rename(field1):

    group = NXgroup(field1)

    assert "f1" in group

    group["f1"].rename("f2")

    assert "f1" not in group
    assert "f2" in group


def test_entry_creation():

    group = NXentry()

    assert group.nxname == "entry"
    assert group.nxclass == "NXentry"
    assert isinstance(group, NXentry)


def test_group_class():

    group = NXgroup()
    group.nxclass = NXentry

    assert group.nxclass == "NXentry"
    assert isinstance(group, NXentry)


def test_group_components():

    g1 = NXdata(name='g1')
    g2 = NXdata(name='g2')
    g3 = NXdata(name='g3')
    group = NXentry(g1, g2, g3)

    assert group.component('NXdata') == [group['g1'], group['g2'], group['g3']]
    assert group.NXdata == [group['g1'], group['g2'], group['g3']]


def test_group_title():

    group = NXentry()
    group["title"] = "Group Title"

    assert group.nxtitle == "Group Title"


def test_group_move(field1):

    group = NXentry()
    group['g1'] = NXgroup()
    group['g1/f1'] = field1
    group['g2'] = NXgroup()
    group['g1'].move('f1', 'g2', name='f2')

    assert 'g1/f1' not in group
    assert 'g2/f2' in group

    group['g2'].move(group['g2/f2'], group['g1'], name='f1')

    assert 'g2/f2' not in group
    assert 'g1/f1' in group

    group['g3'] = NXgroup()
    group['g2/f2'] = NXlink(target='g1/f1')
    group['g2'].move('f2', 'g3', name='f3')

    assert group['g3/f3'].nxlink == field1


def test_group_copy(tmpdir, field1):

    filename = os.path.join(tmpdir, "file1.nxs")
    root = NXroot(NXentry())
    root.save(filename, mode="w")

    external_filename = os.path.join(tmpdir, "file2.nxs")
    external_root = NXroot(
        NXentry(NXgroup(field1, name='g1', attrs={"a": "b"})))
    external_root.save(external_filename, mode="w")

    root["entry/g2"] = NXlink(target="entry/g1", file=external_filename)

    copied_filename = os.path.join(tmpdir, "file3.nxs")
    copied_root = NXroot()
    copied_root.save(copied_filename, mode="w")

    copied_root["entry"] = root["entry"].copy(expand_external=True)

    assert "entry" in copied_root
    assert "g2" in copied_root["entry"]
    assert "entry/g2/f1" in copied_root
    assert not isinstance(copied_root["entry/g2"], NXlink)
    assert copied_root["entry/g2/f1"][0] == 1
    assert "a" in copied_root["entry/g2"].attrs
    assert copied_root["entry/g2"].attrs["a"] == "b"


def test_field_copy(tmpdir, field1):

    filename = os.path.join(tmpdir, "file1.nxs")
    root = NXroot(NXentry(NXgroup(name="g1")))
    root.save(filename, mode="w")

    external_filename = os.path.join(tmpdir, "file2.nxs")
    external_root = NXroot(NXentry(NXgroup(field1, name="g2")))
    external_root.save(external_filename, mode="w")

    root["entry/g1/f1"] = NXlink(target="entry/g2/f1", file=external_filename)

    copied_filename = os.path.join(tmpdir, "file3.nxs")
    copied_root = NXroot(NXentry(NXgroup(name="g3")))
    copied_root.save(copied_filename, mode="w")

    copied_root["entry/g3/f1"] = root["entry/g1/f1"].copy()

    assert "entry/g3/f1" in copied_root
    assert not isinstance(copied_root["entry/g3/f1"], NXlink)
    assert copied_root["entry/g3/f1"][0] == 1



#!/usr/bin/env python
# -----------------------------------------------------------------------------
# Copyright (c) 2019-2021, NeXpy Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING, distributed with this software.
# -----------------------------------------------------------------------------
import argparse

import nexusformat


def main():

    parser = argparse.ArgumentParser(
        description="Determine version number of nexusformat API")
    parser.add_argument('-v', '--version', action='version',
                        version=f'nexusformat v{nexusformat.__version__}')

    parser.parse_args(['--version'])


if __name__ == "__main__":
    main()



# -----------------------------------------------------------------------------
# Copyright (c) 2019-2021, NeXpy Development Team.
#
# Author: Paul Kienzle, Ray Osborn
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING, distributed with this software.
# -----------------------------------------------------------------------------

"""IPython extension to allow autocompletion of NeXus object names.

This modifies h5py.ipy_completer, written by Darren Dale, to accommodate
the completion of NeXus paths defined as nested dictionaries. It will also
autocomplete attributes at the end of a dictionary path. The NeXus objects
can follow an assignment or be embedded in function arguments.

Examples
--------
Autocompletion works on each component of the following commands::

>>> signal = root[entry/data/signal]
>>> units = root[entry/data/signal].units
>>> data = NXdata(root[entry/data/signal])

"""
import re

try:
    import readline
except ImportError:
    readline = None
from string import punctuation

from .tree import NeXusError, NXobject

re_attr_match = re.compile(r"(.+\[.*\].*)\.(\w*)$")
re_item_match = re.compile(r"""(.*)\[(?P<s>['|"])(?!.*(?P=s))(.*)$""")
re_object_match = re.compile(r"(.+?)(?:\[)")


def _retrieve_obj(name, shell):
    """Retrieve the NeXus object at the base of the command.

    This filters out invalid characters not caught by the regex.

    Parameters
    ----------
    name : str
        Name of the object to be retrieved.
    shell : InteractiveShell
        IPython shell containing the namespace to be searched.

    Returns
    -------
    NXobject
        The NeXus object at the base of the command.

    Raises
    ------
    ValueError
        If the object name contains a '('.
    """

    if '(' in name:
        raise ValueError()

    return eval(name, shell.user_ns)


def nxitem_completer(shell, command):
    """Compute possible dictionary matches for NXgroups or NXfields.

    This matches NeXus objects referenced as nested dictionary paths.

    Parameters
    ----------
    shell : InteractiveShell
        IPython shell containing the namespace to be searched.
    command : str
        Command to be autocompleted

    Returns
    -------
    list of str
        List of possible completions.
    """
    base, item = re_item_match.split(command)[1:4:2]

    try:
        obj = _retrieve_obj(base, shell)
    except Exception:
        return []

    import posixpath
    path, _ = posixpath.split(item)
    try:
        if path:
            items = (posixpath.join(path, name) for name in obj[path].keys())
        else:
            items = obj.keys()
    except Exception:
        items = []
    items = list(items)

    readline.set_completer_delims(' \t\n`!@#$^&*()=+[{]}\\|;:\'",<>?')
    return [i for i in items if i[:len(item)] == item]


def nxattr_completer(shell, command):
    """Compute possible matches for NXgroup or NXfield attributes.

    This matches attributes at the end of NeXus dictionary references.
    If the entire NeXus path is defined using attribute references, then
    the autocompletion is handled by other completers.

    Parameters
    ----------
    shell : InteractiveShell
        IPython shell containing the namespace to be searched.
    command : str
        Command to be autocompleted

    Returns
    -------
    list of str
        List of possible completions.
    """
    from IPython import get_ipython
    from IPython.core.error import TryNext
    from IPython.utils import generics
    base, attr = re_attr_match.split(command)[1:3]
    base = base.strip()

    try:
        obj = _retrieve_obj(base, shell)
    except Exception:
        return []

    attrs = obj._get_completion_list()
    try:
        attrs = generics.complete_object(obj, attrs)
    except TryNext:
        pass

    omit__names = get_ipython().Completer.omit__names
    if omit__names == 1:
        attrs = [a for a in attrs if not a.startswith('__')]
    elif omit__names == 2:
        attrs = [a for a in attrs if not a.startswith('_')]

    readline.set_completer_delims(' =')

    return [f".{a}" for a in attrs if a[:len(attr)] == attr]


def nxcompleter(shell, event):
    """Completer function to be loaded into IPython.

    Only text that ends with a valid NXobject is inspected.

    Parameters
    ----------
    shell : InteractiveShell
        IPython shell containing the namespace to be searched.
    event :
        IPython object containing the command to be completed.

    Returns
    -------
    list of str
        List of possible completions.

    Raises
    ------
    TryNext
        If no completions are found.
    """
    from IPython.core.error import TryNext

    if readline is None:
        raise NeXusError(
            "Install the readline module to enable tab completion")

    command = re.split('[ !#$%&()*+,:;<=>?@^~]',
                       event.line)[-1].lstrip(punctuation)
    try:
        base = re_object_match.split(command)[1]
    except Exception:
        raise TryNext

    if not isinstance(shell._ofind(base)['obj'], NXobject):
        raise TryNext

    try:
        return nxattr_completer(shell, command)
    except ValueError:
        pass

    try:
        return nxitem_completer(shell, command)
    except ValueError:
        pass

    return []


def load_ipython_extension(ip=None):
    """Load completer function into IPython.

    This calls the IPython set_hook function to add nxcompleter to the list of
    completer functions. This function disables the use of Jedi autcompletion,
    which is currently incompatible with the nexusformat classes.

    Parameters
    ----------
    ip : InteractiveShell, optional
        IPython shell to be modified. By default, it is set by get_ipython().
    """
    from IPython import get_ipython
    if readline is None:
        raise NeXusError(
            "Install the readline module to enable tab completion")
    if ip is None:
        ip = get_ipython()
    ip.Completer.use_jedi = False
    ip.set_hook('complete_command', nxcompleter,
                re_key=r"(?:.*\=)?(?:.*\()?(?:.*,)?(.+?)\[")
